/*
** Hello Name Server
*/

#include <cbcp.h>
#define CBCP_NET_TCP_IMPLEMENTATION
#include <cbcp_net_tcp.h>

#include <stdio.h>



/*
** CBCP command predeclarations
*/
static void command_add(CBCP_Command_Args *args);


/*
** CBCP program
*/
int main(int argc, char const *argv[]) {
	if (argc < 2) {
		fprintf(stderr, "Usage:\n\t%s <cbcp_database_filepath>\n", argv[0]);
		return 1;
	}

	/*
	** Initialize CBCP by loading the database file for this host
	*/

	const char *cbcp_database_filepath = argv[1];
	CBCP_State *cbcp;

	cbcp_net_tcp_init(0);
	cbcp = cbcp_init(cbcp_database_filepath);

	if (cbcp == NULL) {
		fprintf(stderr, "Could not initialize CBCP.\n");
		return 1;
	}

	/*
	** Bind command callbacks
	*/

	if (!cbcp_server_init_command(
		cbcp,
		"Calculation",
		"Add",
		command_add,
		NULL,
		sizeof(uint32_t))
	) {
		fprintf(stderr, "Could not initialize served command.\n");
		return 1;
	}

	/*
	** Start the cbcp server
	*/

	printf("Starting CBCP server for Hello Name application.\n");

	cbcp_server_start(cbcp);

	return 0;
}

static void command_add(CBCP_Command_Args *args) {
	uint32_t *message = (uint32_t *)args->payload;
	unsigned int message_length = args->payload_length;
	if(message_length != sizeof(uint32_t)*2)
	{
		fprintf(stderr, "message_length is not equal to two 'uint32's!\n"
		                "\tmessage_length: %d\n", message_length);
		return;
	}
	char *response_payload = (char *)args->response_payload;
	unsigned int max_payload_length = args->response_payload_max_length;
	if(max_payload_length < sizeof(uint32_t))
	{
		fprintf(stderr, "max_payload_length is too small!\n");
		return;
	}

	uint32_t a = message[0];
	uint32_t b = message[1];
	uint32_t result = a + b;

	printf(
		"Adding %d and %d. result: %d!\n",
		a, b, result);

	args->response_payload_length = sizeof(uint32_t);
	memcpy(response_payload, (void*)&result, sizeof(result));
}




